@@ -35,6 +35,7 @@
 
 #include "gtksourcelanguage.h"
 #include "gtksourcelanguage-private.h"
+#include "gtksource-marshal.h"
 #include "gtksourceundomanager.h"
 #include "gtksourceundomanagerdefault.h"
 #include "gtksourcestyle.h"
@@ -215,6 +216,7 @@ struct _GtkSourceBufferPrivate
 
 	GtkTextTag *invalid_char_tag;
 
+	guint has_draw_spaces_tag : 1;
 	guint highlight_syntax : 1;
 	guint highlight_brackets : 1;
 	guint implicit_trailing_newline : 1;
@@ -268,10 +270,53 @@ static void	 gtk_source_buffer_real_highlight_updated
 							 GtkTextIter             *start,
 							 GtkTextIter             *end);
 
+static void
+gtk_source_buffer_check_tag_for_spaces (GtkSourceBuffer *buffer,
+                                        GtkSourceTag    *tag)
+{
+	if (!buffer->priv->has_draw_spaces_tag)
+	{
+		gboolean draw_spaces_set;
+
+		g_object_get (tag,
+		              "draw-spaces-set", &draw_spaces_set,
+		              NULL);
+
+		if (draw_spaces_set)
+		{
+			buffer->priv->has_draw_spaces_tag = TRUE;
+		}
+	}
+}
+
+static void
+gtk_source_buffer_tag_changed_cb (GtkTextTagTable *table,
+                                  GtkTextTag      *tag,
+                                  gboolean         size_changed,
+                                  GtkSourceBuffer *buffer)
+{
+	if (GTK_SOURCE_IS_TAG (tag))
+	{
+		gtk_source_buffer_check_tag_for_spaces (buffer, GTK_SOURCE_TAG (tag));
+	}
+}
+
+static void
+gtk_source_buffer_tag_added_cb (GtkTextTagTable *table,
+                                GtkTextTag      *tag,
+                                GtkSourceBuffer *buffer)
+{
+	if (GTK_SOURCE_IS_TAG (tag))
+	{
+		gtk_source_buffer_check_tag_for_spaces (buffer, GTK_SOURCE_TAG (tag));
+	}
+}
+
 static void
 gtk_source_buffer_constructed (GObject *object)
 {
 	GtkSourceBuffer *buffer = GTK_SOURCE_BUFFER (object);
+	GtkTextTagTable *table;
 
 	if (buffer->priv->undo_manager == NULL)
 	{
@@ -280,6 +325,16 @@ gtk_source_buffer_constructed (GObject *object)
 	}
 
 	G_OBJECT_CLASS (gtk_source_buffer_parent_class)->constructed (object);
+
+	table = gtk_text_buffer_get_tag_table (GTK_TEXT_BUFFER (buffer));
+	g_signal_connect_object (table,
+	                         "tag-changed",
+	                         G_CALLBACK (gtk_source_buffer_tag_changed_cb),
+	                         buffer, 0);
+	g_signal_connect_object (table,
+	                         "tag-added",
+	                         G_CALLBACK (gtk_source_buffer_tag_added_cb),
+	                         buffer, 0);
 }
 
 static void
@@ -430,11 +485,15 @@ gtk_source_buffer_class_init (GtkSourceBufferClass *klass)
 	                                G_OBJECT_CLASS_TYPE (object_class),
 	                                G_SIGNAL_RUN_LAST,
 	                                G_CALLBACK (gtk_source_buffer_real_highlight_updated),
-					NULL, NULL, NULL,
+					NULL, NULL,
+					_gtk_source_marshal_VOID__BOXED_BOXED,
 	                                G_TYPE_NONE,
 	                                2,
 	                                GTK_TYPE_TEXT_ITER | G_SIGNAL_TYPE_STATIC_SCOPE,
 	                                GTK_TYPE_TEXT_ITER | G_SIGNAL_TYPE_STATIC_SCOPE);
+	g_signal_set_va_marshaller (buffer_signals[HIGHLIGHT_UPDATED],
+	                            G_TYPE_FROM_CLASS (klass),
+	                            _gtk_source_marshal_VOID__BOXED_BOXEDv);
 
 	/**
 	 * GtkSourceBuffer::source-mark-updated:
@@ -449,9 +508,13 @@ gtk_source_buffer_class_init (GtkSourceBufferClass *klass)
 			   G_OBJECT_CLASS_TYPE (object_class),
 			   G_SIGNAL_RUN_LAST,
 			   0,
-			   NULL, NULL, NULL,
+			   NULL, NULL,
+			   g_cclosure_marshal_VOID__OBJECT,
 			   G_TYPE_NONE,
 			   1, GTK_TYPE_TEXT_MARK);
+	g_signal_set_va_marshaller (buffer_signals[SOURCE_MARK_UPDATED],
+	                            G_TYPE_FROM_CLASS (klass),
+	                            g_cclosure_marshal_VOID__OBJECTv);
 
 	/**
 	 * GtkSourceBuffer::undo:
@@ -465,8 +528,12 @@ gtk_source_buffer_class_init (GtkSourceBufferClass *klass)
 			  G_OBJECT_CLASS_TYPE (object_class),
 			  G_SIGNAL_RUN_LAST,
 			  G_STRUCT_OFFSET (GtkSourceBufferClass, undo),
-			  NULL, NULL, NULL,
+			  NULL, NULL,
+			  g_cclosure_marshal_VOID__VOID,
 			  G_TYPE_NONE, 0);
+	g_signal_set_va_marshaller (buffer_signals[UNDO],
+	                            G_TYPE_FROM_CLASS (klass),
+	                            g_cclosure_marshal_VOID__VOIDv);
 
 	/**
 	 * GtkSourceBuffer::redo:
@@ -479,13 +546,17 @@ gtk_source_buffer_class_init (GtkSourceBufferClass *klass)
 			  G_OBJECT_CLASS_TYPE (object_class),
 			  G_SIGNAL_RUN_LAST,
 			  G_STRUCT_OFFSET (GtkSourceBufferClass, redo),
-			  NULL, NULL, NULL,
+			  NULL, NULL,
+			  g_cclosure_marshal_VOID__VOID,
 			  G_TYPE_NONE, 0);
+	g_signal_set_va_marshaller (buffer_signals[REDO],
+	                            G_TYPE_FROM_CLASS (klass),
+	                            g_cclosure_marshal_VOID__VOIDv);
 
 	/**
 	 * GtkSourceBuffer::bracket-matched:
 	 * @buffer: a #GtkSourceBuffer.
-	 * @iter: if found, the location of the matching bracket.
+	 * @iter: (nullable): if found, the location of the matching bracket.
 	 * @state: state of bracket matching.
 	 *
 	 * @iter is set to a valid iterator pointing to the matching bracket
@@ -504,10 +575,14 @@ gtk_source_buffer_class_init (GtkSourceBufferClass *klass)
 			  G_OBJECT_CLASS_TYPE (object_class),
 			  G_SIGNAL_RUN_LAST,
 			  G_STRUCT_OFFSET (GtkSourceBufferClass, bracket_matched),
-			  NULL, NULL, NULL,
+			  NULL, NULL,
+			  _gtk_source_marshal_VOID__BOXED_ENUM,
 			  G_TYPE_NONE, 2,
 			  GTK_TYPE_TEXT_ITER,
 			  GTK_SOURCE_TYPE_BRACKET_MATCH_TYPE);
+	g_signal_set_va_marshaller (buffer_signals[BRACKET_MATCHED],
+	                            G_TYPE_FROM_CLASS (klass),
+	                            _gtk_source_marshal_VOID__BOXED_ENUMv);
 }
 
 static void
@@ -2051,6 +2126,12 @@ gtk_source_buffer_create_source_mark (GtkSourceBuffer   *buffer,
 				  GTK_TEXT_MARK (mark),
 				  where);
 
+	/* We want to return a borrowed reference and the mark is already
+	 * owned by @buffer due to gtk_text_buffer_add_mark(). Therefore
+	 * it is safe to unref immediately.
+	 */
+	g_object_unref (mark);
+
 	return mark;
 }
 
@@ -2110,7 +2191,7 @@ _gtk_source_buffer_source_mark_prev (GtkSourceBuffer *buffer,
 /**
  * gtk_source_buffer_forward_iter_to_source_mark:
  * @buffer: a #GtkSourceBuffer.
- * @iter: an iterator.
+ * @iter: (inout): an iterator.
  * @category: (nullable): category to search for, or %NULL
  *
  * Moves @iter to the position of the next #GtkSourceMark of the given
@@ -2144,7 +2225,7 @@ gtk_source_buffer_forward_iter_to_source_mark (GtkSourceBuffer *buffer,
 /**
  * gtk_source_buffer_backward_iter_to_source_mark:
  * @buffer: a #GtkSourceBuffer.
- * @iter: an iterator.
+ * @iter: (inout): an iterator.
  * @category: (nullable): category to search for, or %NULL
  *
  * Moves @iter to the position of the previous #GtkSourceMark of the given
@@ -2298,6 +2379,19 @@ gtk_source_buffer_remove_source_marks (GtkSourceBuffer   *buffer,
 	g_slist_free (list);
 }
 
+gboolean
+_gtk_source_buffer_has_source_marks (GtkSourceBuffer *buffer)
+{
+	g_return_val_if_fail (GTK_SOURCE_IS_BUFFER (buffer), FALSE);
+
+	if (buffer->priv->all_source_marks != NULL)
+	{
+		return !_gtk_source_marks_sequence_is_empty (buffer->priv->all_source_marks);
+	}
+
+	return FALSE;
+}
+
 static GtkTextTag *
 get_context_class_tag (GtkSourceBuffer *buffer,
 		       const gchar     *context_class)
@@ -2406,7 +2500,7 @@ gtk_source_buffer_get_context_classes_at_iter (GtkSourceBuffer   *buffer,
 /**
  * gtk_source_buffer_iter_forward_to_context_class_toggle:
  * @buffer: a #GtkSourceBuffer.
- * @iter: a #GtkTextIter.
+ * @iter: (inout): a #GtkTextIter.
  * @context_class: the context class.
  *
  * Moves forward to the next toggle (on or off) of the context class. If no
@@ -2445,7 +2539,7 @@ gtk_source_buffer_iter_forward_to_context_class_toggle (GtkSourceBuffer *buffer,
 /**
  * gtk_source_buffer_iter_backward_to_context_class_toggle:
  * @buffer: a #GtkSourceBuffer.
- * @iter: a #GtkTextIter.
+ * @iter: (inout): a #GtkTextIter.
  * @context_class: the context class.
  *
  * Moves backward to the next toggle (on or off) of the context class. If no
@@ -3309,3 +3403,11 @@ gtk_source_buffer_create_source_tag (GtkSourceBuffer *buffer,
 
 	return tag;
 }
+
+gboolean
+_gtk_source_buffer_has_spaces_tag (GtkSourceBuffer *buffer)
+{
+	g_return_val_if_fail (GTK_SOURCE_IS_BUFFER (buffer), FALSE);
+
+	return buffer->priv->has_draw_spaces_tag;
+}
